# coding: utf-8

# docs
from typing import List, NamedTuple, Dict, Optional


from supervisely.api.module_api import ApiField
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.pointcloud_annotation.pointcloud_annotation import PointcloudAnnotation
from supervisely.pointcloud_annotation.pointcloud_object_collection import PointcloudObjectCollection
from supervisely.api.entity_annotation.entity_annotation_api import EntityAnnotationAPI


class PointcloudAnnotationAPI(EntityAnnotationAPI):
    _method_download_bulk = 'point-clouds.annotations.bulk.info'
    _entity_ids_str = ApiField.POINTCLOUD_IDS

    def download(self, pointcloud_id: int) -> List[Dict]:
        """
        :param pointcloud_id: int
        :return: list of content of pointcloud object with given id, received after execution post request
        """
        info = self._api.pointcloud.get_info_by_id(pointcloud_id)
        return self._download(info.dataset_id, pointcloud_id)

    def append(self, pointcloud_id: int, ann: PointcloudAnnotation, key_id_map: Optional[KeyIdMap] = None) -> None:
        info = self._api.pointcloud.get_info_by_id(pointcloud_id)

        new_objects = []
        for object_3d in ann.objects:
            if key_id_map is not None and key_id_map.get_object_id(object_3d.key()) is not None:
                # object already uploaded
                continue
            new_objects.append(object_3d)

        self._append(self._api.pointcloud.tag, self._api.pointcloud.object, self._api.pointcloud.figure,
                     info.project_id, info.dataset_id, pointcloud_id,
                     ann.tags, PointcloudObjectCollection(new_objects), ann.figures, key_id_map)



from typing import Dict, Optional

from supervisely.app import StateJson
from supervisely.app.widgets import Widget

class RandomSplitsTable(Widget):
    def __init__(
        self, 
        items_count: int, 
        start_train_percent: Optional[int] = 80,
        widget_id: Optional[int] = None
    ):
        if 1 <= start_train_percent <= 99:
            pass
        else:
            raise ValueError("start_train_percent must be in range [1; 99].")
        self._table_data = [
            {"name": "train", "type": "success"},
            {"name": "val", "type": "primary"},
            {"name": "total", "type": "gray"},
        ]
        self._items_count = items_count
        train_count = int(items_count / 100 * start_train_percent)
        self._count = {
            "total": items_count,
            "train": train_count,
            "val": items_count - train_count
        }

        self._percent = {
            "total": 100,
            "train": start_train_percent,
            "val": 100 - start_train_percent
        }

        super().__init__(widget_id=widget_id, file_path=__file__)


    def get_json_data(self):
        return {
            "table_data": self._table_data,
            "items_count": self._items_count,
        }

    def get_json_state(self):
        return {
            "count": self._count,
            "percent": self._percent
        }

    def get_splits_counts(self) -> Dict[str, int]:
        return StateJson()[self.widget_id]["count"]


# coding: utf-8


import os
import json
from typing import List, Union
import numpy as np

import pydicom
import stringcase
from supervisely.io.fs import get_file_ext, list_files_recursively, list_files
import supervisely.volume.nrrd_encoder as nrrd_encoder
from supervisely import logger

# Do NOT use directly for extension validation. Use is_valid_ext() /  has_valid_ext() below instead.
ALLOWED_VOLUME_EXTENSIONS = [".nrrd", ".dcm"]


class UnsupportedVolumeFormat(Exception):
    pass


def get_extension(path: str):
    # magic.from_file("path", mime=True)
    # for nrrd:
    # application/octet-stream
    # for nifti(nii):
    # application/octet-stream
    # for dicom:
    # "application/dicom"

    ext = get_file_ext(path)
    if ext in ALLOWED_VOLUME_EXTENSIONS:
        return ext

    # case when dicom file does not have an extension
    import magic

    mime = magic.from_file(path, mime=True)
    if mime == "application/dicom":
        return ".dcm"
    return None


def is_valid_ext(ext: str) -> bool:
    return ext.lower() in ALLOWED_VOLUME_EXTENSIONS


def has_valid_ext(path: str) -> bool:
    return is_valid_ext(get_extension(path))


def validate_format(path: str):
    if not has_valid_ext(path):
        raise UnsupportedVolumeFormat(
            f"File {path} has unsupported volume extension. Supported extensions: {ALLOWED_VOLUME_EXTENSIONS}"
        )


def rescale_slope_intercept(value, slope, intercept):
    return value * slope + intercept


def normalize_volume_meta(meta):
    meta["intensity"]["min"] = rescale_slope_intercept(
        meta["intensity"]["min"],
        meta["rescaleSlope"],
        meta["rescaleIntercept"],
    )

    meta["intensity"]["max"] = rescale_slope_intercept(
        meta["intensity"]["max"],
        meta["rescaleSlope"],
        meta["rescaleIntercept"],
    )

    if "windowWidth" not in meta:
        meta["windowWidth"] = meta["intensity"]["max"] - meta["intensity"]["min"]

    if "windowCenter" not in meta:
        meta["windowCenter"] = meta["intensity"]["min"] + meta["windowWidth"] / 2

    return meta


def read_dicom_serie_volume_np(paths: List[str], anonymize=True) -> np.ndarray:
    import SimpleITK as sitk
    sitk_volume, meta = read_dicom_serie_volume(paths, anonymize=anonymize)
    # for debug:
    # sitk.WriteImage(sitk_volume, "/work/output/sitk.nrrd", useCompression=False, compressionLevel=9)
    # with open("/work/output/test.nrrd", "wb") as file:
    #     file.write(b)
    volume_np = sitk.GetArrayFromImage(sitk_volume)
    volume_np = np.transpose(volume_np, (2, 1, 0))
    return volume_np, meta


_anonymize_tags = ["PatientID", "PatientName"]
_default_dicom_tags = [
    "SeriesInstanceUID",
    "Modality",
    "WindowCenter",
    "WindowWidth",
    "RescaleIntercept",
    "RescaleSlope",
    "PhotometricInterpretation",
]
_default_dicom_tags.extend(_anonymize_tags)

_photometricInterpretationRGB = set(
    [
        "RGB",
        "PALETTE COLOR",
        "YBR_FULL",
        "YBR_FULL_422",
        "YBR_PARTIAL_422",
        "YBR_PARTIAL_420",
        "YBR_RCT",
    ]
)


def read_dicom_tags(
    path, allowed_keys: Union[None, List[str]] = _default_dicom_tags, anonymize=True
):
    import SimpleITK as sitk
    reader = sitk.ImageFileReader()
    reader.SetFileName(path)
    reader.LoadPrivateTagsOn()
    reader.ReadImageInformation()

    vol_info = {}
    for k in reader.GetMetaDataKeys():
        v = reader.GetMetaData(k)
        tag = pydicom.tag.Tag(k.split("|")[0], k.split("|")[1])
        keyword = pydicom.datadict.keyword_for_tag(tag)
        if allowed_keys is not None and keyword not in allowed_keys:
            continue
        if anonymize is True and keyword in _anonymize_tags:
            v = "anonymized"
        keyword = stringcase.camelcase(keyword)
        vol_info[keyword] = v
        if keyword in [
            "windowCenter",
            "windowWidth",
            "rescaleIntercept",
            "rescaleSlope",
        ]:
            vol_info[keyword] = float(vol_info[keyword].split("\\")[0])
        elif (
            keyword == "photometricInterpretation"
            and v in _photometricInterpretationRGB
        ):
            vol_info["channelsCount"] = 3
    return vol_info


def encode(volume_np: np.ndarray, volume_meta):
    directions = np.array(volume_meta["directions"]).reshape(3, 3)
    directions *= volume_meta["spacing"]

    volume_bytes = nrrd_encoder.encode(
        volume_np,
        header={
            "encoding": "gzip",
            # "space": "left-posterior-superior",
            "space": "right-anterior-superior",
            "space directions": directions.T.tolist(),
            "space origin": volume_meta["origin"],
        },
        compression_level=1,
    )

    # with open("/work/output/test.nrrd", "wb") as file:
    #     file.write(volume_bytes)

    return volume_bytes


def inspect_dicom_series(root_dir: str):
    import SimpleITK as sitk
    found_series = {}
    for d in os.walk(root_dir):
        dir = d[0]
        reader = sitk.ImageSeriesReader()
        sitk.ProcessObject_SetGlobalWarningDisplay(False)
        series_found = reader.GetGDCMSeriesIDs(dir)
        sitk.ProcessObject_SetGlobalWarningDisplay(True)
        logger.info(f"Found {len(series_found)} series in directory {dir}")
        for serie in series_found:
            dicom_names = reader.GetGDCMSeriesFileNames(dir, serie)
            found_series[serie] = dicom_names
    logger.info(f"Total {len(found_series)} series in directory {root_dir}")
    return found_series


def _sitk_image_orient_ras(sitk_volume):
    import SimpleITK as sitk
    if sitk_volume.GetDimension() == 4 and sitk_volume.GetSize()[3] == 1:
        sitk_volume = sitk_volume[:, :, :, 0]

    sitk_volume = sitk.DICOMOrient(sitk_volume, "RAS")
    # RAS reorient image using filter
    # orientation_filter = sitk.DICOMOrientImageFilter()
    # orientation_filter.SetDesiredCoordinateOrientation("RAS")
    # sitk_volume = orientation_filter.Execute(sitk_volume)

    # https://discourse.itk.org/t/getdirection-and-getorigin-for-simpleitk-c-implementation/3472/8
    origin = list(sitk_volume.GetOrigin())
    directions = list(sitk_volume.GetDirection())
    origin[0] *= -1
    origin[1] *= -1
    directions[0] *= -1
    directions[1] *= -1
    directions[3] *= -1
    directions[4] *= -1
    directions[6] *= -1
    directions[7] *= -1
    sitk_volume.SetOrigin(origin)
    sitk_volume.SetDirection(directions)
    return sitk_volume


def read_dicom_serie_volume(paths, anonymize=True):
    import SimpleITK as sitk
    reader = sitk.ImageSeriesReader()
    reader.SetFileNames(paths)
    sitk_volume = reader.Execute()
    sitk_volume = _sitk_image_orient_ras(sitk_volume)
    dicom_tags = read_dicom_tags(paths[0], anonymize=anonymize)

    f_min_max = sitk.MinimumMaximumImageFilter()
    f_min_max.Execute(sitk_volume)
    meta = get_meta(
        sitk_volume.GetSize(),
        f_min_max.GetMinimum(),
        f_min_max.GetMaximum(),
        sitk_volume.GetSpacing(),
        sitk_volume.GetOrigin(),
        sitk_volume.GetDirection(),
        dicom_tags,
    )
    return sitk_volume, meta


def compose_ijk_2_world_mat(spacing, origin, directions):
    mat = np.eye(4)
    mat[:3, :3] = (np.array(directions).reshape(3, 3) * spacing).T
    mat[:3, 3] = origin
    return mat


def get_meta(
    sitk_shape, min_intensity, max_intensity, spacing, origin, directions, dicom_tags={}
):

    # x = 1 - sagittal
    # y = 1 - coronal
    # z = 1 - axial
    volume_meta = normalize_volume_meta(
        {
            **dicom_tags,
            "channelsCount": 1,
            "rescaleSlope": 1,
            "rescaleIntercept": 0,
            "intensity": {
                "min": min_intensity,
                "max": max_intensity,
            },
            "dimensionsIJK": {
                "x": sitk_shape[0],
                "y": sitk_shape[1],
                "z": sitk_shape[2],
            },
            "ACS": "RAS",
            # instead of IJK2WorldMatrix field
            "spacing": spacing,
            "origin": origin,
            "directions": directions,
        }
    )
    return volume_meta


def inspect_nrrd_series(root_dir: str):
    nrrd_paths = list_files_recursively(root_dir, [".nrrd"])
    logger.info(f"Total {len(nrrd_paths)} nnrd series in directory {root_dir}")
    return nrrd_paths


def read_nrrd_serie_volume(path: str):
    import SimpleITK as sitk
    # find custom NRRD loader in gitlab supervisely_py/-/blob/feature/import-volumes/plugins/import/volumes/src/loaders/nrrd.py
    reader = sitk.ImageFileReader()
    # reader.SetImageIO("NrrdImageIO")
    reader.SetFileName(path)
    sitk_volume = reader.Execute()

    sitk_volume = _sitk_image_orient_ras(sitk_volume)
    f_min_max = sitk.MinimumMaximumImageFilter()
    f_min_max.Execute(sitk_volume)
    meta = get_meta(
        sitk_volume.GetSize(),
        f_min_max.GetMinimum(),
        f_min_max.GetMaximum(),
        sitk_volume.GetSpacing(),
        sitk_volume.GetOrigin(),
        sitk_volume.GetDirection(),
        {},
    )
    return sitk_volume, meta


def read_nrrd_serie_volume_np(paths: List[str]) -> np.ndarray:
    import SimpleITK as sitk
    sitk_volume, meta = read_nrrd_serie_volume(paths)
    volume_np = sitk.GetArrayFromImage(sitk_volume)
    volume_np = np.transpose(volume_np, (2, 1, 0))
    return volume_np, meta



from __future__ import annotations
from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget, ConditionalWidget, ConditionalItem
from typing import List, Dict, Union


class Switch(ConditionalWidget):
    class Routes:
        VALUE_CHANGED = "value_changed"

    def __init__(
            self,
            switched: bool = False,
            width: int = 58,  # default value 46 for empty text
            on_text: str = "ON",
            off_text: str = "OFF",
            on_color: str = None,  # default: "#20a0ff"
            off_color: str = None,  # default: "#bfcbd9"
            on_content: Widget = None,
            off_content: Widget = None,
            widget_id: str = None,
    ):
        self._switched = switched
        self._width = width
        self._on_text = on_text
        self._off_text = off_text
        self._on_color = on_color
        self._off_color = off_color
        self._changes_handled = False
        items = [
            ConditionalItem(value="1", label="1", content=on_content),
            ConditionalItem(value="0", label="0", content=off_content)
        ]
        super().__init__(items=items, widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        return {
            "width": self._width,
            "onText": self._on_text,
            "offText": self._off_text,
            "onColor": self._on_color,
            "offColor": self._off_color
        }

    def get_json_state(self) -> Dict:
        return {"switched": self._switched, "value": str(int(self._switched))}

    def is_switched(self):
        return StateJson()[self.widget_id]["switched"]

    def on(self):
        StateJson()[self.widget_id]["switched"] = True
        StateJson()[self.widget_id]["value"] = "1"
        StateJson().send_changes()

    def off(self):
        StateJson()[self.widget_id]["switched"] = False
        StateJson()[self.widget_id]["value"] = "0"
        StateJson().send_changes()

    def get_width(self):
        return DataJson()[self.widget_id]["width"]

    def get_on_text(self):
        return DataJson()[self.widget_id]["onText"]

    def set_on_text(self, value: str):
        DataJson()[self.widget_id]["onText"] = value
        DataJson().send_changes()

    def get_off_text(self):
        return DataJson()[self.widget_id]["offText"]

    def set_off_text(self, value: str):
        DataJson()[self.widget_id]["offText"] = value
        DataJson().send_changes()

    def get_on_color(self):
        return DataJson()[self.widget_id]["onColor"]

    def set_on_color(self, value: str):
        DataJson()[self.widget_id]["onColor"] = value
        DataJson().send_changes()

    def get_off_color(self):
        return DataJson()[self.widget_id]["offColor"]

    def set_off_color(self, value: str):
        DataJson()[self.widget_id]["offColor"] = value
        DataJson().send_changes()

    def value_changed(self, func):
        route_path = self.get_route_path(Switch.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            res = self.is_switched()
            StateJson()[self.widget_id]["value"] = str(int(res))
            StateJson().send_changes()
            func(res)

        return _click



# coding: utf-8
"""augmentations for images and annotations"""

from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import imgaug.augmenters.Sequential
from typing import Tuple, List, Dict, Optional

import random
import numpy as np

from supervisely.imaging import image as sly_image
from supervisely.annotation.annotation import Annotation
from supervisely.geometry.image_rotator import ImageRotator
from supervisely.geometry.rectangle import Rectangle
from supervisely._utils import take_with_default
from supervisely.sly_logger import logger


def _validate_image_annotation_shape(img: np.ndarray, ann: Annotation) -> None:
    if img.shape[:2] != ann.img_size:
        raise RuntimeError(
            "Image shape {} doesn't match img_size {} in annotation.".format(
                img.shape[:2], ann.img_size
            )
        )


# Flips
def fliplr(img: np.ndarray, ann: Annotation) -> Tuple[np.ndarray, Annotation]:
    """
    Flips an Image and Annotation around vertical axis.

    :param img: Image in numpy format, :class:`RGB`.
    :type img: np.ndarray
    :param ann: Annotation object.
    :type ann: Annotation
    :raises: :class:`RuntimeError` if Image shape does not match img_size in Annotation
    :return: Tuple containing flipped Image and Annotation
    :rtype: :class:`Tuple[np.ndarray, Annotation]`
    :Usage Example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.aug.aug import fliplr

        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Download image and annotation from API
        project_id = 116501
        image_id = 193940171

        meta_json = api.project.get_meta(project_id)
        meta = sly.ProjectMeta.from_json(meta_json)

        image_np = api.image.download_np(image_id) # <class 'numpy.ndarray'>
        ann_info = api.annotation.download(image_id)
        ann = sly.Annotation.from_json(ann_info.annotation, meta)

        # Flip image and annotation
        flip_image_np, flip_ann = fliplr(image_np, ann)
    """
    _validate_image_annotation_shape(img, ann)
    res_img = sly_image.fliplr(img)
    res_ann = ann.fliplr()
    return res_img, res_ann


def flipud(img: np.ndarray, ann: Annotation) -> Tuple[np.ndarray, Annotation]:
    """
    Flips an Image and Annotation around horizontal axis.

    :param img: Image in numpy format, :class:`RGB`.
    :type img: np.ndarray
    :param ann: Annotation object.
    :type ann: Annotation
    :raises: :class:`RuntimeError` if Image shape does not match img_size in Annotation
    :return: Tuple containing flipped Image and Annotation
    :rtype: :class:`Tuple[np.ndarray, Annotation]`
    :Usage Example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.aug.aug import flipud

        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Download image and annotation from API
        project_id = 116501
        image_id = 193940171

        meta_json = api.project.get_meta(project_id)
        meta = sly.ProjectMeta.from_json(meta_json)

        image_np = api.image.download_np(image_id) # <class 'numpy.ndarray'>
        ann_info = api.annotation.download(image_id)
        ann = sly.Annotation.from_json(ann_info.annotation, meta)

        # Flip image and annotation
        flip_image_np, flip_ann = flipud(image_np, ann)
    """
    _validate_image_annotation_shape(img, ann)
    res_img = sly_image.flipud(img)
    res_ann = ann.flipud()
    return res_img, res_ann


# Crops
def crop(
    img: np.ndarray,
    ann: Annotation,
    top_pad: Optional[int] = 0,
    left_pad: Optional[int] = 0,
    bottom_pad: Optional[int] = 0,
    right_pad: Optional[int] = 0,
) -> Tuple[np.ndarray, Annotation]:
    """
    Crops an Image and Annotation from all sides with a given values.

    :param img: Image in numpy format, :class:`RGB`.
    :type img: np.ndarray
    :param ann: Annotation object.
    :type ann: Annotation
    :param top_pad: Top padding in pixels.
    :type top_pad: int, optional
    :param left_pad: Left padding in pixels.
    :type left_pad: int, optional
    :param bottom_pad: Bottom padding in pixels.
    :type bottom_pad: int, optional
    :param right_pad: Right padding in pixels.
    :type right_pad: int, optional
    :raises: :class:`RuntimeError` if Image shape does not match img_size in Annotation
    :return: Tuple containing cropped Image and Annotation
    :rtype: :class:`Tuple[np.ndarray, Annotation]`
    :Usage Example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.aug.aug import crop

        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Download image and annotation from API
        project_id = 116501
        image_id = 193940171

        meta_json = api.project.get_meta(project_id)
        meta = sly.ProjectMeta.from_json(meta_json)

        image_np = api.image.download_np(image_id) # <class 'numpy.ndarray'>
        print(image_np.shape)
        # Output: (800, 1067, 3)

        ann_info = api.annotation.download(image_id)
        ann = sly.Annotation.from_json(ann_info.annotation, meta)

        crop_image_np, crop_ann = crop(image_np, ann, top_pad=50, left_pad=100, bottom_pad=50, right_pad=100)
        print(crop_image_np.shape)
        # Output: (700, 867, 3)
    """
    _validate_image_annotation_shape(img, ann)
    height, width = img.shape[:2]
    crop_rect = Rectangle(
        top_pad, left_pad, height - bottom_pad - 1, width - right_pad - 1
    )

    res_img = sly_image.crop(img, crop_rect)
    res_ann = ann.relative_crop(crop_rect)
    return res_img, res_ann


def crop_fraction(
    img: np.ndarray,
    ann: Annotation,
    top: Optional[float] = 0,
    left: Optional[float] = 0,
    bottom: Optional[float] = 0,
    right: Optional[float] = 0,
) -> Tuple[np.ndarray, Annotation]:
    """
    Crops an Image and Annotation from all sides with the given fraction values.

    :param img: Image in numpy format, :class:`RGB`.
    :type img: np.ndarray
    :param ann: Annotation object.
    :type ann: Annotation
    :param top: Top padding in pixels.
    :type top: int, optional
    :param left: Left padding in pixels.
    :type left: int, optional
    :param bottom: Bottom padding in pixels.
    :type bottom: int, optional
    :param right: Right padding in pixels.
    :type right: int, optional
    :raises: :class:`ValueError` if fraction values not between 0 and 1
    :return: Tuple containing cropped Image and Annotation
    :rtype: :class:`Tuple[np.ndarray, Annotation]`
    :Usage Example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.aug.aug import crop_fraction

        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Download image and annotation from API
        project_id = 116501
        image_id = 193940171

        meta_json = api.project.get_meta(project_id)
        meta = sly.ProjectMeta.from_json(meta_json)

        image_np = api.image.download_np(image_id) # <class 'numpy.ndarray'>
        print(image_np.shape)
        # Output: (800, 1067, 3)

        ann_info = api.annotation.download(image_id)
        ann = sly.Annotation.from_json(ann_info.annotation, meta)

        crop_image_np, crop_ann = crop_fraction(image_np, ann, top=0.1, left=0.2, bottom=0.1, right=0.2)
        print(crop_image_np.shape)
        # Output: (640, 641, 3)
    """
    _validate_image_annotation_shape(img, ann)
    if not all(0 <= pad < 1 for pad in (top, left, right, bottom)):
        raise ValueError("All padding values must be between 0 and 1.")
    height, width = img.shape[:2]
    top_pixels = round(height * top)
    left_pixels = round(width * left)
    bottom_pixels = round(height * bottom)
    right_pixels = round(width * right)
    return crop(
        img,
        ann,
        top_pad=top_pixels,
        left_pad=left_pixels,
        bottom_pad=bottom_pixels,
        right_pad=right_pixels,
    )


def random_crop(
    img: np.ndarray, ann: Annotation, height: int, width: int
) -> Tuple[np.ndarray, Annotation]:
    """
    Crops an Image and Annotation at a random location.

    :param img: Image in numpy format, :class:`RGB`.
    :type img: np.ndarray
    :param ann: Annotation object.
    :type ann: Annotation
    :param height: Desired height of output crop.
    :type height: int, optional
    :param width: Desired width of output crop.
    :type width: int, optional
    :raises: :class:`RuntimeError` if Image shape does not match img_size in Annotation
    :return: Tuple containing cropped Image and Annotation
    :rtype: :class:`Tuple[np.ndarray, Annotation]`
    :Usage Example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.aug.aug import random_crop

        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Download image and annotation from API
        project_id = 116501
        image_id = 193940171

        meta_json = api.project.get_meta(project_id)
        meta = sly.ProjectMeta.from_json(meta_json)

        image_np = api.image.download_np(image_id) # <class 'numpy.ndarray'>
        print(image_np.shape)
        # Output: (800, 1067, 3)

        ann_info = api.annotation.download(image_id)
        ann = sly.Annotation.from_json(ann_info.annotation, meta)

        crop_image_np, crop_ann = random_crop(image_np, ann, height=500, width=700)
        print(crop_image_np.shape)
        # Output: (500, 700, 3)
    """
    _validate_image_annotation_shape(img, ann)
    img_height, img_width = img.shape[:2]

    def calc_crop_pad(old_side, crop_side):
        new_side = min(int(old_side), int(crop_side))
        min_bound = random.randint(0, old_side - new_side)  # including [a; b]
        max_bound = old_side - min_bound - new_side
        return min_bound, max_bound

    left_pad, right_pad = calc_crop_pad(img_width, width)
    top_pad, bottom_pad = calc_crop_pad(img_height, height)
    return crop(
        img,
        ann,
        top_pad=top_pad,
        left_pad=left_pad,
        bottom_pad=bottom_pad,
        right_pad=right_pad,
    )


def random_crop_fraction(
    img: np.ndarray,
    ann: Annotation,
    height_fraction_range: Tuple,
    width_fraction_range: Tuple,
) -> Tuple[np.ndarray, Annotation]:
    """
    Crops an Image and Annotation at a random location with random size in a given interval.

    :param img: Image in numpy format, :class:`RGB`.
    :type img: np.ndarray
    :param ann: Annotation object.
    :type ann: Annotation
    :param height_fraction_range: Range of relative values [0, 1] to select output height from.
    :type height_fraction_range: Tuple[float, float]
    :param width_fraction_range: Range of relative values [0, 1] to select output width from.
    :type width_fraction_range: Tuple[float, float]
    :raises: :class:`RuntimeError` if Image shape does not match img_size in Annotation
    :return: Tuple containing cropped Image and Annotation
    :rtype: :class:`Tuple[np.ndarray, Annotation]`
    :Usage Example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.aug.aug import random_crop_fraction

        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Download image and annotation from API
        project_id = 116501
        image_id = 193940171

        meta_json = api.project.get_meta(project_id)
        meta = sly.ProjectMeta.from_json(meta_json)

        image_np = api.image.download_np(image_id) # <class 'numpy.ndarray'>
        print(image_np.shape)
        # Output: (800, 1067, 3)

        ann_info = api.annotation.download(image_id)
        ann = sly.Annotation.from_json(ann_info.annotation, meta)

        crop_image_np, crop_ann = random_crop_fraction(image_np, ann, height_fraction_range=(0.1, 0.8), width_fraction_range=(0.1, 0.8))
        print(crop_image_np.shape)
        # Output: (486, 585, 3)
    """
    _validate_image_annotation_shape(img, ann)
    img_height, img_width = img.shape[:2]

    height_p = random.uniform(height_fraction_range[0], height_fraction_range[1])
    width_p = random.uniform(width_fraction_range[0], width_fraction_range[1])
    crop_height = round(img_height * height_p)
    crop_width = round(img_width * width_p)
    return random_crop(img, ann, height=crop_height, width=crop_width)


def batch_random_crops_fraction(
    img_ann_pairs: List[Tuple[np.ndarray, Annotation]],
    crops_per_image: int,
    height_fraction_range: Tuple,
    width_fraction_range: Tuple,
) -> List[Tuple[np.ndarray, Annotation]]:
    return [
        random_crop_fraction(img, ann, height_fraction_range, width_fraction_range)
        for img, ann in img_ann_pairs
        for _ in range(crops_per_image)
    ]


def flip_add_random_crops(
    img: np.ndarray,
    ann: Annotation,
    crops_per_image: int,
    height_fraction_range: Tuple,
    width_fraction_range: Tuple,
) -> List[Tuple[np.ndarray, Annotation]]:
    full_size_items = [(img, ann), fliplr(img, ann)]
    crops = batch_random_crops_fraction(
        full_size_items, crops_per_image, height_fraction_range, width_fraction_range
    )
    return full_size_items + crops


# TODO factor out / simplify.
def _rect_from_bounds(padding_config: Dict, img_h: int, img_w: int) -> Rectangle:
    def get_padding_pixels(raw_side, dim_name):
        side_padding_config = padding_config.get(dim_name)
        if side_padding_config is None:
            padding_pixels = 0
        elif side_padding_config.endswith("px"):
            padding_pixels = int(side_padding_config[: -len("px")])
        elif side_padding_config.endswith("%"):
            padding_fraction = float(side_padding_config[: -len("%")])
            padding_pixels = int(raw_side * padding_fraction / 100.0)
        else:
            raise ValueError(
                'Unknown padding size format: {}. Expected absolute values as "5px" or relative as "5%"'.format(
                    side_padding_config
                )
            )
        return padding_pixels

    def get_padded_side(raw_side, l_name, r_name):
        l_bound = -get_padding_pixels(raw_side, l_name)
        r_bound = raw_side + get_padding_pixels(raw_side, r_name)
        return l_bound, r_bound

    left, right = get_padded_side(img_w, "left", "right")
    top, bottom = get_padded_side(img_h, "top", "bottom")
    return Rectangle(top=top, left=left, bottom=bottom, right=right)


def instance_crop(
    img: np.ndarray,
    ann: Annotation,
    class_title: str,
    save_other_classes_in_crop: Optional[bool] = True,
    padding_config: Optional[Dict[str, str]] = None,
) -> List[Tuple[np.ndarray, Annotation]]:
    """
    Crops objects of specified classes from Image and Annotation with configurable padding.

    :param img: Image in numpy format, :class:`RGB`.
    :type img: np.ndarray
    :param ann: Annotation object.
    :type ann: Annotation
    :param class_title: Name of class to crop.
    :type class_title: str
    :param save_other_classes_in_crop: If True saves non-target classes in each cropped Annotation, otherwise don't.
    :type save_other_classes_in_crop: bool, optional
    :param padding_config: Dict with padding.
    :type padding_config: dict, optional
    :raises: :class:`ValueError` if padding size format is incorrect
    :return: List of cropped (image numpy array, Annotation) pairs
    :rtype: :class:`List[Tuple[np.ndarray, Annotation]]`

    :Usage Example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.aug.aug import instance_crop

        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Download image and annotation from API
        project_id = 116501
        image_id = 193940171

        meta_json = api.project.get_meta(project_id)
        meta = sly.ProjectMeta.from_json(meta_json)

        image_np = api.image.download_np(image_id) # <class 'numpy.ndarray'>
        print(image_np.shape)
        # Output: (800, 1067, 3)

        ann_info = api.annotation.download(image_id)
        ann = sly.Annotation.from_json(ann_info.annotation, meta)

        result = instance_crop(image_np, ann, 'kiwi', True, {'top': '20px', 'left': '50px', 'bottom': '700px', 'right': '1000px'})
        for crop_image_np, crop_ann in result:
            print(crop_image_np.shape)
            # Output: (270, 635, 3)
            #         (426, 345, 3)
    """
    padding_config = take_with_default(padding_config, {})
    _validate_image_annotation_shape(img, ann)
    results = []
    img_rect = Rectangle.from_size(img.shape[:2])

    if save_other_classes_in_crop:
        non_target_labels = [
            label for label in ann.labels if label.obj_class.name != class_title
        ]
    else:
        non_target_labels = []

    ann_with_non_target_labels = ann.clone(labels=non_target_labels)

    for label in ann.labels:
        if label.obj_class.name == class_title:
            src_fig_rect = label.geometry.to_bbox()
            new_img_rect = _rect_from_bounds(
                padding_config, img_w=src_fig_rect.width, img_h=src_fig_rect.height
            )
            rect_to_crop = new_img_rect.translate(src_fig_rect.top, src_fig_rect.left)
            crops = rect_to_crop.crop(img_rect)
            if len(crops) == 0:
                continue
            rect_to_crop = crops[0]
            image_crop = sly_image.crop(img, rect_to_crop)

            cropped_ann = ann_with_non_target_labels.relative_crop(rect_to_crop)

            label_crops = label.relative_crop(rect_to_crop)
            for label_crop in label_crops:
                results.append((image_crop, cropped_ann.add_label(label_crop)))
    return results


# Resize
def resize(
    img: np.ndarray, ann: Annotation, size: Tuple
) -> Tuple[np.ndarray, Annotation]:
    """
    Resizes an input Image and Annotation to a given size.

    :param img: Image in numpy format, :class:`RGB`.
    :type img: np.ndarray
    :param ann: Annotation object.
    :type ann: Annotation
    :param size: Desired size (height, width) in pixels or -1.
    :type size: Tuple[int, int]
    :raises: :class:`RuntimeError` if Image shape does not match img_size in Annotation
    :return: Tuple containing resized Image and Annotation
    :rtype: :class:`Tuple[np.ndarray, Annotation]`

    :Usage Example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.aug.aug import resize

        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Download image and annotation from API
        project_id = 116501
        image_id = 193940171

        meta_json = api.project.get_meta(project_id)
        meta = sly.ProjectMeta.from_json(meta_json)

        image_np = api.image.download_np(image_id) # <class 'numpy.ndarray'>
        print(image_np.shape)
        # Output: (800, 1067, 3)

        ann_info = api.annotation.download(image_id)
        ann = sly.Annotation.from_json(ann_info.annotation, meta)

        resize_image_np, resize_ann = resize(image_np, ann, (600, -1))
        print(resize_image_np.shape)
        # Output: (600, 800, 3)
    """
    _validate_image_annotation_shape(img, ann)
    height = take_with_default(size[0], -1)  # For backward capability
    width = take_with_default(size[1], -1)
    size = (height, width)

    new_size = sly_image.restore_proportional_size(in_size=ann.img_size, out_size=size)
    res_img = sly_image.resize(img, new_size)
    res_ann = ann.resize(new_size)
    return res_img, res_ann


# Resize
def scale(
    img: np.ndarray,
    ann: Annotation,
    frow: Optional[float] = None,
    fcol: Optional[float] = None,
    f: Optional[float] = None,
) -> Tuple[np.ndarray, Annotation]:
    """
    Scales an input Image and Annotation to a given size.

    :param img: Image in numpy format, :class:`RGB`.
    :type img: np.ndarray
    :param ann: Annotation object.
    :type ann: Annotation
    :param frow: Desired height scale height value.
    :type frow: float, optional
    :param fcol: Desired width scale height value.
    :type fcol: float, optional
    :param f: Desired height and width scale values in one(positive).
    :type f: float, optional
    :raises: :class:`RuntimeError` if Image shape does not match img_size in Annotation
    :return: Tuple containing scaled Image and Annotation
    :rtype: :class:`Tuple[np.ndarray, Annotation]`

    :Usage Example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.aug.aug import scale

        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Download image and annotation from API
        project_id = 116501
        image_id = 193940171

        meta_json = api.project.get_meta(project_id)
        meta = sly.ProjectMeta.from_json(meta_json)

        image_np = api.image.download_np(image_id) # <class 'numpy.ndarray'>
        print(image_np.shape)
        # Output: (800, 1067, 3)

        ann_info = api.annotation.download(193940171)
        ann = sly.Annotation.from_json(ann_info.annotation, meta)

        scale_image_np, scale_ann = scale(image_np, ann, frow=0.7, fcol=0.8)
        print(scale_image_np.shape)
        # Output: (560, 854, 3)
    """
    _validate_image_annotation_shape(img, ann)
    new_size = sly_image.restore_proportional_size(
        in_size=ann.img_size, frow=frow, fcol=fcol, f=f
    )
    res_img = sly_image.resize(img, new_size)
    res_ann = ann.resize(new_size)
    return res_img, res_ann


# Rotate
class RotationModes:
    KEEP = "keep"
    CROP = "crop"


def rotate(
    img: np.ndarray,
    ann: Annotation,
    degrees: float,
    mode: Optional[str] = RotationModes.KEEP,
) -> Tuple[np.ndarray, Annotation]:  # @TODO: add "preserve_size" mode
    """
    Rotates an Image and Annotation by random angle.

    :param img: Image in numpy format, :class:`RGB`.
    :type img: np.ndarray
    :param ann: Annotation object.
    :type ann: Annotation
    :param degrees: Rotation angle.
    :type degrees: int
    :param mode: One of RotateMode enum values.
    :type mode: RotationModes, optional
    :raises: :class:`RuntimeError` if Image shape does not match img_size in Annotation
    :return: Tuple containing rotated Image and Annotation
    :rtype: :class:`Tuple[np.ndarray, Annotation]`

    :Usage Example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.aug.aug import rotate

        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Download image and annotation from API
        project_id = 116501
        image_id = 193940171

        meta_json = api.project.get_meta(project_id)
        meta = sly.ProjectMeta.from_json(meta_json)

        image_np = api.image.download_np(image_id) # <class 'numpy.ndarray'>
        print(image_np.shape)
        # Output: (800, 1067, 3)

        ann_info = api.annotation.download(image_id)
        ann = sly.Annotation.from_json(ann_info.annotation, meta)

        rotate_image_np, rotate_ann = rotate(image_np, ann, 30)
        print(rotate_image_np.shape)
        # Output: (1231, 1326, 3)
    """
    _validate_image_annotation_shape(img, ann)
    rotator = ImageRotator(img.shape[:2], degrees)

    if mode == RotationModes.KEEP:
        rect_to_crop = None

    elif mode == RotationModes.CROP:
        rect_to_crop = rotator.inner_crop

    else:
        raise NotImplementedError("Wrong black_regions mode.")

    res_img = rotator.rotate_img(img, use_inter_nearest=False)
    res_ann = ann.rotate(rotator)
    if rect_to_crop is not None:
        res_img = sly_image.crop(res_img, rect_to_crop)
        res_ann = res_ann.relative_crop(rect_to_crop)
    return res_img, res_ann


def load_imgaug(json_data: Dict) -> imgaug.augmenters.Sequential:
    import imgaug.augmenters as iaa

    def _get_function(category_name, aug_name):
        try:
            submodule = getattr(iaa, category_name)
            aug_f = getattr(submodule, aug_name)
            return aug_f
        except Exception as e:
            logger.error(repr(e))
            raise e

    pipeline_json = json_data["pipeline"]
    random_order = json_data.get("random_order", False)

    pipeline = []
    for aug_info in pipeline_json:
        category_name = aug_info["category"]
        aug_name = aug_info["name"]
        params = aug_info["params"]

        aug_func = _get_function(category_name, aug_name)
        aug = aug_func(**params)

        sometimes = aug_info.get("sometimes", None)
        if sometimes is not None:
            aug = iaa.meta.Sometimes(sometimes, aug)
        pipeline.append(aug)
    augs = iaa.Sequential(pipeline, random_order=random_order)
    return augs



# coding: utf-8
from __future__ import annotations

import random
import colorsys
import os
import gzip
import json
import copy
from typing import List


def _validate_color(color):
    """
    Checks input color for compliance with the required format
    :param: color: color (RGB tuple of integers)
    """
    if not isinstance(color, (list, tuple)):
        raise ValueError("Color has to be list, or tuple")
    if len(color) != 3:
        raise ValueError("Color have to contain exactly 3 values: [R, G, B]")
    for channel in color:
        validate_channel_value(channel)


def random_rgb() -> List[int, int, int]:
    """
    Generate RGB color with fixed saturation and lightness.

    :return: RGB integer values
    :rtype: :class:`List[int, int, int]`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        color = sly.color.random_rgb()
        print(color)
        # Output: [138, 15, 123]
    """
    hsl_color = (random.random(), 0.3, 0.8)
    rgb_color = colorsys.hls_to_rgb(*hsl_color)
    return [round(c * 255) for c in rgb_color]


def _normalize_color(color):
    """
    Divide all RGB values by 255.
    :param color: color (RGB tuple of integers)
    """
    return [c / 255.0 for c in color]


def _color_distance(first_color: list, second_color: list) -> float:
    """
    Calculate distance in HLS color space between Hue components of 2 colors
    :param first_color: first color (RGB tuple of integers)
    :param second_color: second color (RGB tuple of integers)
    :return: Euclidean distance between 'first_color' and 'second_color'
    """
    first_color_hls = colorsys.rgb_to_hls(*_normalize_color(first_color))
    second_color_hls = colorsys.rgb_to_hls(*_normalize_color(second_color))
    hue_distance = min(
        abs(first_color_hls[0] - second_color_hls[0]),
        1 - abs(first_color_hls[0] - second_color_hls[0]),
    )
    return hue_distance


def generate_rgb(exist_colors: List[List[int, int, int]]) -> List[int, int, int]:
    """
    Generate new color which oppositely by exist colors.

    :param exist_colors: List of existing colors in RGB format.
    :type exist_colors: list
    :return: RGB integer values
    :rtype: :class:`List[int, int, int]`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        exist_colors = [[0, 0, 0], [128, 64, 255]]
        color = sly.color.generate_rgb(exist_colors)
        print(color)
        # Output: [15, 138, 39]
    """
    largest_min_distance = 0
    best_color = random_rgb()
    if len(exist_colors) > 0:
        for _ in range(100):
            color = random_rgb()
            current_min_distance = min(_color_distance(color, c) for c in exist_colors)
            if current_min_distance > largest_min_distance:
                largest_min_distance = current_min_distance
                best_color = color
    _validate_color(best_color)
    return best_color


def rgb2hex(color: List[int, int, int]) -> str:
    """
    Convert integer color format to HEX string.

    :param color: List of existing colors in RGB format.
    :type color: List[int, int, int]
    :return: HEX RGB string
    :rtype: :class:`str`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        hex_color = sly.color.rgb2hex([128, 64, 255])
        print(hex_color)
        # Output: #8040FF
    """
    _validate_color(color)
    return "#" + "".join("{:02X}".format(component) for component in color)


def _hex2color(hex_value: str) -> list:
    """
    Convert HEX RGB string to integer RGB format
    :param hex_value: HEX RGBA string. Example: "#FF02А4
    :return: RGB integer values. Example: [80, 255, 0]
    """
    assert hex_value.startswith("#")
    return [int(hex_value[i : (i + 2)], 16) for i in range(1, len(hex_value), 2)]


def hex2rgb(hex_value: str) -> List[int, int, int]:
    """
    Convert HEX RGB string to integer RGB format.

    :param hex_value: HEX RGB string.
    :type hex_value: str
    :return: RGB integer values
    :rtype: :class:`List[int, int, int]`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        hex_color = '#8040FF'
        color = sly.color.hex2rgb(hex_color)
        print(color)
        # Output: [128, 64, 255]
    """
    assert len(hex_value) == 7, "Supported only HEX RGB string format!"
    color = _hex2color(hex_value)
    _validate_color(color)
    return color


def _hex2rgba(hex_value: str) -> list:
    """
    Convert HEX RGBA string to integer RGBA format
    :param hex_value: HEX RGBA string. Example: "#FF02А4CC
    :return: RGBA integer values. Example: [80, 255, 0, 128]
    """
    assert len(hex_value) == 9, "Supported only HEX RGBA string format!"
    return _hex2color(hex_value)


def validate_channel_value(value: int) -> None:
    """
    Generates ValueError if value not between 0 and 255.

    :param value: Input channel value.
    :type value: int
    :raises: :class:`ValueError` if value not between 0 and 255.
    :return: None
    :rtype: :class:`NoneType`
    """
    if 0 <= value <= 255:
        pass
    else:
        raise ValueError("Color channel has to be in range [0; 255]")


# generate colors
# import distinctipy
# data = {}
# for n in range(100):
#     print(n)
#     colors = distinctipy.get_colors(n)
#     rgb_colors = [distinctipy.get_rgb256(color) for color in colors]
#     data[n] = rgb_colors
# sly.json.dump_json_file(data, "colors.json")

# import gzip
# data = sly.json.load_json_file("colors.json")
# with gzip.open("colors.json.gz", "wt", encoding="UTF-8") as zipfile:
#     json.dump(data, zipfile)
# with gzip.open("colors.json.gz", "r") as fin:
#     data = json.loads(fin.read().decode("utf-8"))


def get_predefined_colors(n: int):
    try:
        file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "colors.json.gz")
        with gzip.open(file, "r") as fin:
            data = json.loads(fin.read().decode("utf-8"))

        if str(n) in data:
            colors = copy.deepcopy(data[str(n)])
            random.Random(7).shuffle(colors)
            return colors
        # generate random colors
    except Exception as e:
        print(repr(e))

    rand_colors = []
    for i in range(n):
        rand_colors.append(random_rgb())
    return rand_colors



# coding: utf-8

import os
import numpy as np
from typing import List, Optional
from supervisely._utils import is_development, abs_url
from supervisely.io.fs import ensure_base_path

# Do NOT use directly for extension validation. Use is_valid_ext() /  has_valid_ext() below instead.
ALLOWED_POINTCLOUD_EXTENSIONS = [".pcd"]


class PointcloudExtensionError(Exception):
    pass


class UnsupportedPointcloudFormat(Exception):
    pass


class PointcloudReadException(Exception):
    pass


def is_valid_ext(ext: str) -> bool:
    """
    Checks if given extention is supported
    :param ext: str
    :return: bool
    """
    return ext.lower() in ALLOWED_POINTCLOUD_EXTENSIONS


def has_valid_ext(path: str) -> bool:
    """
    Checks if file from given path with given extention is supported
    :param path: str
    :return: bool
    """
    return is_valid_ext(os.path.splitext(path)[1])


def validate_ext(ext: str) -> None:
    """
    Raise error if given extention is not supported
    :param ext: str
    """
    if not is_valid_ext(ext):
        raise UnsupportedPointcloudFormat(
            "Unsupported pointcloud extension: {}. Only the following extensions are supported: {}.".format(
                ext, ALLOWED_POINTCLOUD_EXTENSIONS
            )
        )


def validate_format(path):
    _, ext = os.path.splitext(path)
    validate_ext(ext)


def get_labeling_tool_url(dataset_id, pointcloud_id):
    res = f"/app/point-clouds/?datasetId={dataset_id}&pointCloudId={pointcloud_id}"
    if is_development():
        res = abs_url(res)
    return res


def get_labeling_tool_link(url, name="open in labeling tool"):
    return f'<a href="{url}" rel="noopener noreferrer" target="_blank">{name}<i class="zmdi zmdi-open-in-new" style="margin-left: 5px"></i></a>'


def read(path: str, coords_dims: Optional[List[int]] = None) -> np.ndarray:
    """
    Loads a pointcloud from the specified file and returns it in XYZ format.

    :param path: Path to file.
    :type path: str
    :return: Numpy array
    :rtype: :class:`np.ndarray`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        ptc = sly.pointcloud.read('/home/admin/work/pointclouds/ptc0.pcd')
    """
    try:
        import open3d as o3d
    except ImportError:
        raise ImportError("No module named open3d. Please make sure that module is installed from pip and try again.")
    validate_format(path)
    if coords_dims is None:
        coords_dims = [0, 1, 2]
    pcd_data = o3d.io.read_point_cloud(path)
    if pcd_data is None:
        raise IOError(f"open3d can not open the file {path}")
    pointcloud_np = np.asarray(pcd_data.points)
    pointcloud_np = pointcloud_np[:, coords_dims]
    return pointcloud_np


def write(path: str, pointcloud_np: np.ndarray, coords_dims: Optional[List[int]] = None) -> bool:
    """
    Saves a pointcloud to the specified file. It creates directory from path if the directory for this path does not exist.

    :param path: Path to file.
    :type path: str
    :param pointcloud_np: Pointcloud [N, 3] in XYZ format.
    :type pointcloud_np: :class:`np.ndarray`
    :param coords_dims: List of indexes for (X, Y, Z) coords. Default (if None): [0, 1, 2].
    :type coords_dims: Optional[List[int]]
    :return: Success or not.
    :rtype: bool
    :Usage example:

     .. code-block:: python

        import supervisely as sly
        import numpy as np

        pointcloud = np.random.randn(100, 3)

        ptc = sly.pointcloud.write('/home/admin/work/pointclouds/ptc0.pcd', pointcloud)
    """
    try:
        import open3d as o3d
    except ImportError:
        raise ImportError("No module named open3d. Please make sure that module is installed from pip and try again.")
    ensure_base_path(path)
    validate_format(path)
    if coords_dims is None:
        coords_dims = [0, 1, 2]
    pointcloud_np = pointcloud_np[:, coords_dims]
    pcd_data = o3d.geometry.PointCloud()
    pcd_data.points = o3d.utility.Vector3dVector(pointcloud_np)
    return o3d.io.write_point_cloud(path, pcd_data)



# coding: utf-8

import logging
import types
import datetime
import os
from collections import namedtuple
from enum import Enum
#import simplejson
from pythonjsonlogger import jsonlogger


###############################################################################

class ServiceType(Enum):
    AGENT = 1
    TASK = 2
    EXPORT = 3


class EventType(Enum):
    LOGJ = 1
    LOGS = 2
    TASK_STARTED = 3
    TASK_FINISHED = 4
    TASK_STOPPED = 5
    TASK_CRASHED = 6
    PROGRESS = 7
    CHECKPOINT = 8
    METRICS = 9
    TASK_REMOVED = 10
    MODEL_APPLIED = 11
    DTL_APPLIED = 12
    IMPORT_APPLIED = 13
    PROJECT_CREATED = 14
    TASK_VERIFIED = 15
    STEP_COMPLETE = 16
    TASK_DEPLOYED = 17
    AGENT_READY_FOR_TASKS = 18
    MISSED_TASK_FOUND = 19
    REPORT_CREATED = 20
    APP_FINISHED = 21


###############################################################################
# predefined levels


# level name: level, default exc_info, description
LogLevelSpec = namedtuple('LogLevelSpec', [
    'int',
    'add_exc_info',
    'descr',
])


LOGGING_LEVELS = {
    'FATAL': LogLevelSpec(50, True, 'Critical error'),
    'ERROR': LogLevelSpec(40, True, 'Error'),   # may be shown to end user
    'WARN': LogLevelSpec(30, False, 'Warning'),   # may be shown to end user
    'INFO': LogLevelSpec(20, False, 'Info'),   # may be shown to end user
    'DEBUG': LogLevelSpec(10, False, 'Debug'),
    'TRACE': LogLevelSpec(5, False, 'Trace'),
}


def _set_logging_levels(levels, the_logger):
    for lvl_name, (lvl, def_exc_info, _) in levels.items():
        logging.addLevelName(lvl, lvl_name.upper())  # two mappings

        def construct_logger_member(lvl_val, default_exc_info):
            return lambda self, msg, *args, exc_info=default_exc_info, **kwargs: \
                self.log(lvl_val,
                         msg,
                         *args,
                         exc_info=exc_info,
                         **kwargs)

        func = construct_logger_member(lvl, def_exc_info)
        bound_method = types.MethodType(func, the_logger)
        setattr(the_logger, lvl_name.lower(), bound_method)


###############################################################################


def _get_default_logging_fields():
    supported_keys = [
        'asctime',
        # 'created',
        # 'filename',
        # 'funcName',
        'levelname',
        # 'levelno',
        # 'lineno',
        # 'module',
        # 'msecs',
        'message',
        # 'name',
        # 'pathname',
        # 'process',
        # 'processName',
        # 'relativeCreated',
        # 'thread',
        # 'threadName'
    ]
    return ' '.join(['%({0:s})'.format(k) for k in supported_keys])


#def dumps_ignore_nan(obj, *args, **kwargs):
#    return  simplejson.dumps(obj, ignore_nan=True, *args, **kwargs)


class CustomJsonFormatter(jsonlogger.JsonFormatter):
    additional_fields = {}

    def __init__(self, format_string):
        super().__init__(format_string)#, json_serializer=dumps_ignore_nan)

    def process_log_record(self, log_record):
        log_record['timestamp'] = log_record.pop('asctime', None)

        levelname = log_record.pop('levelname', None)
        if levelname is not None:
            log_record['level'] = levelname.lower()

        e_info = log_record.pop('exc_info', None)
        if e_info is not None:
            if e_info == 'NoneType: None':  # python logger is not ok here
                pass
            else:
                log_record['stack'] = e_info.split('\n')

        return jsonlogger.JsonFormatter.process_log_record(self, log_record)

    def add_fields(self, log_record, record, message_dict):
        super(CustomJsonFormatter, self).add_fields(log_record, record, message_dict)

        for field, val in CustomJsonFormatter.additional_fields.items():
            if (val is not None) and (field not in log_record):
                log_record[field] = val

    def formatTime(self, record, datefmt=None):
        ct = datetime.datetime.fromtimestamp(record.created)
        t = ct.strftime('%Y-%m-%dT%H:%M:%S')
        s = '%s.%03dZ' % (t, record.msecs)
        return s


def _construct_logger(the_logger, loglevel_text):
    for handler in the_logger.handlers:
        the_logger.removeHandler(handler)

    _set_logging_levels(LOGGING_LEVELS, the_logger)

    the_logger.setLevel(loglevel_text.upper())

    log_handler = logging.StreamHandler()
    add_logger_handler(the_logger, log_handler)

    the_logger.propagate = False


###############################################################################


def add_logger_handler(the_logger, log_handler):  # default format
    logger_fmt_string = _get_default_logging_fields()
    formatter = CustomJsonFormatter(logger_fmt_string)
    log_handler.setFormatter(formatter)
    the_logger.addHandler(log_handler)


def add_default_logging_into_file(the_logger, log_dir):
    fname = 'log_{}.txt'.format(
        datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S"))
    ofpath = os.path.join(log_dir, fname)

    log_handler_file = logging.FileHandler(filename=ofpath)
    add_logger_handler(the_logger, log_handler_file)


# runs on all formatters
def change_formatters_default_values(the_logger, field_name, value):
    for handler in the_logger.handlers:
        hfaf = handler.formatter.additional_fields
        if value is not None:
            hfaf[field_name] = value
        else:
            hfaf.pop(field_name, None)


def _get_loglevel_env():
    loglevel = os.getenv('LOG_LEVEL', None)
    if loglevel is None:
        loglevel = os.getenv('LOGLEVEL', 'INFO')
    return loglevel.upper()


def set_global_logger():
    loglevel = _get_loglevel_env()  # use the env to set loglevel
    the_logger = logging.getLogger('logger')  # optional logger name
    _construct_logger(the_logger, loglevel)
    return the_logger


def get_task_logger(task_id, loglevel=None):
    if loglevel is None:
        loglevel = _get_loglevel_env() # use the env to set loglevel
    logger_name = 'task_{}'.format(task_id)
    the_logger = logging.getLogger(logger_name)  # optional logger name
    _construct_logger(the_logger, loglevel)
    return the_logger


logger = set_global_logger()

